orx_linked_list

Type Alias DoublyIdx

source
pub type DoublyIdx<T> = NodeIdx<Doubly<T>>;
An index to an element on a doubly linked list which allows safe and constant time access.

Aliased Type

struct DoublyIdx<T> { /* private fields */ }