orx_linked_list

Type Alias SinglyIdx

source
pub type SinglyIdx<T> = NodeIdx<Singly<T>>;
An index to an element on a singly linked list which allows safe and constant time access.

Aliased Type

struct SinglyIdx<T> { /* private fields */ }